#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <time.h>
#include <fcntl.h>
#include <X11/Xatom.h>

#include "frm.xpm"
#include "mask.xbm"
#include "lcd.xpm"

#define VERSION "0.4"
#define APM_FILE "/proc/apm"

int ICONIFIED=0; /* default is not iconified */
/* with shape */
#define LED_W 10
#define LED_H 15
#define LED_AC_X LED_W
#define LED_AC_Y (2*LED_H)
#define LED_AC_W (2*LED_W)
#define LED_AC_H LED_H
#define LED_CHARGE_W 4
#define LED_CHARGE_H 7
#define LED_PERCENT_W 8
#define LED_PERCENT_H 6
#define LED_BATTERY_W 14
#define LED_BATTERY_H 13
#define LED_BLINK_W 1
#define LED_BLINK_H 3
int ledx[] = {
 0,LED_W,2*LED_W,3*LED_W,4*LED_W,    0,LED_W,2*LED_W,3*LED_W,4*LED_W,      0};
int ledy[] = {
 0,    0,      0,      0,      0,LED_H,LED_H,  LED_H,  LED_H,  LED_H,2*LED_H};
int chargex[] = {0,        LED_CHARGE_W};
int chargey[] = {(3*LED_H),  (3*LED_H)};
int percentx[] = {(2*LED_CHARGE_W), (2*LED_CHARGE_W+LED_PERCENT_W)};
int percenty[] = {(3*LED_H), (3*LED_H)};
int batteryx[] = {0, LED_BATTERY_W};
int batteryy[] = {(3*LED_H+LED_CHARGE_H), (3*LED_H+LED_CHARGE_H)};
int blinkx[] = {
(2*LED_CHARGE_W+2*LED_PERCENT_W), (2*LED_CHARGE_W+2*LED_PERCENT_W+1)};
int blinky[] = {(3*LED_H), (3*LED_H)};

/* position */
int posx[] = {5, LED_W+5, 2*LED_W+5+2, 3*LED_W+5+2};
int posy[] = {6,       6,           6,           6};
#define POS_PERCENT_X (4*LED_W+5+2)
#define POS_PERCENT_Y 15
#define POS_BATTERY_X (4*LED_W+5+3+LED_PERCENT_W)
#define POS_BATTERY_Y 7
#define POS_CHARGE_X (4*LED_W+5+4+LED_PERCENT_W+LED_BATTERY_W)
#define POS_CHARGE_Y 10
#define POS_BLINK_X (posx[2]-2)
#define POS_BLINK_Y 12
void charge_disp();
void charge_nodisp();
void ac_disp();
void ac_nodisp();
void percent_disp();
void percent_nodisp();
void battery_disp();
void battery_nodisp();
void blink_disp();
void blink_nodisp();

/* X11 Variablen *****************************************************/
Display *dpy;    /* welches DISPLAY */
Window Root;      /* Hintergrund-Drawable */
int screen;
int x_fd;
int d_depth;
XSizeHints mysizehints;
XWMHints mywmhints;
Pixel back_pix, fore_pix;
GC NormalGC;
Window iconwin, win;       /* My home is my window */
char *ProgName;
char *Geometry;
char Execute[] = "echo no program has been specified >/dev/console";
char *ERR_colorcells = "not enough free color cells\n";
char *ampers = " &";
/* XPM Variablen *****************************************************/
typedef struct _XpmIcon {
    Pixmap pixmap;
    Pixmap mask;
    XpmAttributes attributes;
}        XpmIcon;

XpmIcon pc110lcd, led;
XpmIcon visible;
time_t actualtime;
long actualmin;
int mode = 0;

/* lokale Funktionen *************************************************/
#define MW_EVENTS   (ExposureMask | ButtonPressMask | StructureNotifyMask)
#define FALSE 0
#define TRUE 1
void GetXPM(void);
static void CreatePixmap(Display *dpy, Window Root, char **data, XpmIcon* icon);
Pixel GetColor(char *name);
void RedrawWindow( XpmIcon *v);
void InitLCD();
void InsertLCD();
void InsertLCD2();
/*********************************************************************/
/*********************************************************************/
int updatespeed = 5;
static char *help_message[] = {
"where options include:",
"  -u <secs>              : updatespeed",
"  -exe <program>         : program to start on click",
"  -position [+|-]x[+|-]y : position of pc110lcd",
"  -iconic                : start up as icon",
"  -1                     : clock mode",
"  -3                     : clock & battery mode",
NULL
};

void usage()
{
  char **cpp;

  fprintf(stderr,"pc110lcd ver %s\n",VERSION);
  fprintf(stderr,"usage:  %s [-options ...] \n", ProgName);
  for (cpp = help_message; *cpp; cpp++) {
    fprintf(stderr, "%s\n", *cpp);
  }
  fprintf(stderr,"\n");
  exit(1);
}
int main(int argc,char* argv[])
{
  int i, flag = FALSE;
  unsigned int borderwidth ;
  char *display_name = NULL; 
  char *wname = "pc110lcd";
  XGCValues gcv;
  unsigned long gcm;
  XEvent Event;
  XTextProperty name;
  XClassHint classHint;
  Pixmap pixmask;
  ProgName = argv[0];
  Geometry = "";
  
  /* Parse command line options */
  ProgName = argv[0];

  for(i=1;i<argc;i++) {
    char *arg= argv[i];

    if (arg[0] == '-') {
      switch(arg[1]) {
      case 'u':
        if (++i >= argc) usage();
        sscanf(argv[i],"%d",&updatespeed);
        continue;
      case 'e':
        if(++i >=argc) usage();
        strcpy(&Execute[0], argv[i]);
        strcat(&Execute[0], " &");
        continue;
      case 'p':
        if(++i >=argc) usage();
        Geometry = argv[i];
        continue;
      case 'i':
        ICONIFIED=1;
        continue;
      case '1':
        mode=1;
        continue;
      case '3':
        mode=3;
        continue;
      default:
        usage();
      }
    }
  }

  /* Open the display */
  if (!(dpy = XOpenDisplay(display_name)))  { 
      fprintf(stderr,"pc110lcd: can't open display %s\n", 
        XDisplayName(display_name)); 
      exit (1); 
  } 
  screen= DefaultScreen(dpy);
  Root = RootWindow(dpy, screen);
  d_depth = DefaultDepth(dpy, screen);
  x_fd = XConnectionNumber(dpy);
  
  /* Icon Daten nach XImage konvertieren */
  GetXPM();
  
  /* Create a window to hold the banner */
  mysizehints.flags= USSize|USPosition;
  mysizehints.x = 0;
  mysizehints.y = 0;

  back_pix = GetColor("white");
  fore_pix = GetColor("black");

  XWMGeometry(dpy, screen, Geometry, NULL, (borderwidth =1), &mysizehints,
    &mysizehints.x,&mysizehints.y,&mysizehints.width,&mysizehints.height, &i); 

  mysizehints.width = pc110lcd.attributes.width;
  mysizehints.height= pc110lcd.attributes.height;

  win = XCreateSimpleWindow(dpy,Root,mysizehints.x,mysizehints.y,
          mysizehints.width,mysizehints.height,
          borderwidth,fore_pix,back_pix);
  iconwin = XCreateSimpleWindow(dpy,win,mysizehints.x,mysizehints.y,
              mysizehints.width,mysizehints.height,
              borderwidth,fore_pix,back_pix);



  /* Hints aktivieren */
  XSetWMNormalHints(dpy, win, &mysizehints);
  classHint.res_name =  "pc110lcd";
  classHint.res_class = "PC110LCD";
  XSetClassHint(dpy, win, &classHint);

  XSelectInput(dpy,win,MW_EVENTS);
  XSelectInput(dpy,iconwin,MW_EVENTS);
  
  if (XStringListToTextProperty(&wname, 1, &name) ==0) {
    fprintf(stderr, "pc110lcd: can't allocate window name\n");
    exit(-1);
  }
  XSetWMName(dpy, win, &name);
  
  /* Create a GC for drawing */
  gcm = GCForeground|GCBackground|GCGraphicsExposures;
  gcv.foreground = fore_pix;
  gcv.background = back_pix;
  gcv.graphics_exposures = FALSE;
  NormalGC = XCreateGC(dpy, Root, gcm, &gcv);  

    pixmask = XCreateBitmapFromData(dpy, win, mask_bits, mask_width, 
            mask_height);
    XShapeCombineMask(dpy, win, ShapeBounding, 0, 0, pixmask, ShapeSet);
    XShapeCombineMask(dpy, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
  
  mywmhints.initial_state = WithdrawnState;
  mywmhints.icon_window = iconwin;
  mywmhints.icon_x = mysizehints.x;
  mywmhints.icon_y = mysizehints.y;
  mywmhints.window_group = win;
  mywmhints.flags = StateHint | IconWindowHint | IconPositionHint 
      | WindowGroupHint;
  XSetWMHints(dpy, win, &mywmhints); 

  XSetCommand(dpy, win, argv, argc);
  XMapWindow(dpy,win);

  InitLCD();
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[mode], ledy[mode], LED_W, LED_H, posx[0], posy[0]);

  RedrawWindow(&visible);
  while(1) {
    if (actualtime != time(0)) {
      actualtime = time(0);
      if(actualtime % updatespeed == 0) {
        switch (mode) {
        case 0:
        case 2:
          InsertLCD();
          break;
        case 1:
          InsertLCD();
          InsertLCD2();
          flag = TRUE;
          break;
        case 3:
          if (flag) {
            InsertLCD();
            flag = FALSE;
          } else {
            InsertLCD2();
            flag = TRUE;
          }
          break;
        }
      }
      if ((mode == 1 || mode == 3) && flag) {
        if (actualtime%2)
          blink_disp();
        else
          blink_nodisp();
      }
      RedrawWindow(&visible);
    }
      
    /* read a packet */
    while (XPending(dpy)) {
      XNextEvent(dpy,&Event);
      switch(Event.type) {
      case Expose:
        if(Event.xexpose.count == 0 )
          RedrawWindow(&visible);
        break;
      case ButtonPress:
        system(Execute);
        break;
      case DestroyNotify:
      /*
       * Don't need these as XCloseDisplay() will
       * free them. 
        XFreeGC(dpy, NormalGC);
        XDestroyWindow(dpy, win);
        XDestroyWindow(dpy, iconwin);
      */
        XCloseDisplay(dpy);
        exit(0); 
      default:
        break;      
      }
    }
    XFlush(dpy);
#ifdef SYSV
    poll((struct poll *) 0, (size_t) 0, 50);
#else
    usleep(50000L);      /* 5/100 sec */
#endif
  }
  return 0;
}
/****************************************************************************/
void nocolor(char *a, char *b)
{
  fprintf(stderr,"pc110lcd: can't %s %s\n", a,b);
}
/****************************************************************************/
/* Konvertiere XPMIcons nach XImage */
void GetXPM(void)
{
  XColor col;
  XWindowAttributes attributes;
  int ret;

  /* for the colormap */
  XGetWindowAttributes(dpy,Root,&attributes);

  CreatePixmap(dpy, Root, frm_xpm, &pc110lcd);
  CreatePixmap(dpy, Root, frm_xpm, &visible);
  CreatePixmap(dpy, Root, lcd_xpm, &led);
}
/****************************************************************************/
static void CreatePixmap(Display *dpy, Window Root, char **data, XpmIcon* icon)
{
  int ret;

  icon->attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions
         | XpmExactColors | XpmCloseness);
  icon->attributes.exactColors=False;
  icon->attributes.closeness=40000;
  ret = XpmCreatePixmapFromData(dpy, Root, data, &(icon->pixmap), 
        &(icon->mask), &(icon->attributes));
  if(ret != XpmSuccess)
    {fprintf(stderr, ERR_colorcells);exit(1);}
}
/****************************************************************************/
/* Removes expose events for a specific window from the queue */
int flush_expose (Window w)
{
  XEvent dummy;
  int i=0;
  
  while (XCheckTypedWindowEvent (dpy, w, Expose, &dummy))i++;
  return i;
}

/****************************************************************************/
/* Draws the icon window */
void RedrawWindow( XpmIcon *v)
{
  flush_expose (iconwin);
  XCopyArea(dpy,v->pixmap,iconwin,NormalGC,
      0,0,v->attributes.width, v->attributes.height,0,0);
  flush_expose (win);
  XCopyArea(dpy,v->pixmap,win,NormalGC,
      0,0,v->attributes.width, v->attributes.height,0,0);

}
/****************************************************************************/
Pixel GetColor(char *name)
{
  XColor color;
  XWindowAttributes attributes;

  XGetWindowAttributes(dpy,Root,&attributes);
  color.pixel = 0;
  if (!XParseColor (dpy, attributes.colormap, name, &color)) {
     nocolor("parse",name);
  } else if(!XAllocColor (dpy, attributes.colormap, &color)) {
    nocolor("alloc",name);
  }
  return color.pixel;
}
/*********************************************************************/
void InitLCD()
{
  XCopyArea(dpy, pc110lcd.pixmap, visible.pixmap, NormalGC,
    0, 0, mysizehints.width, mysizehints.height, 0, 0);
}
void charge_on()
{
    XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
      chargex[0], chargey[0], LED_CHARGE_W, LED_CHARGE_H, POS_CHARGE_X, POS_CHARGE_Y);
}
void charge_off()
{
    XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
      chargex[1], chargey[1], LED_CHARGE_W, LED_CHARGE_H, POS_CHARGE_X, POS_CHARGE_Y);
}
void percent_disp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    percentx[0], percenty[0], LED_PERCENT_W, LED_PERCENT_H, POS_PERCENT_X, POS_PERCENT_Y);
}
void percent_nodisp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    percentx[1], percenty[1], LED_PERCENT_W, LED_PERCENT_H, POS_PERCENT_X, POS_PERCENT_Y);
}
void ac_disp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    LED_AC_X, LED_AC_Y, LED_AC_W, LED_AC_H, posx[2], posy[2]);
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[10], ledy[10], LED_W, LED_H, posx[1], posy[1]);
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[10], ledy[10], LED_W, LED_H, posx[0], posy[0]);
  percent_nodisp();
}
void battery_disp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    batteryx[0], batteryy[0], LED_BATTERY_W, LED_BATTERY_H, POS_BATTERY_X, POS_BATTERY_Y);
}
void battery_nodisp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    batteryx[1], batteryy[1], LED_BATTERY_W, LED_BATTERY_H, POS_BATTERY_X, POS_BATTERY_Y);
}
void blink_disp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    blinkx[0], blinky[0], LED_BLINK_W, LED_BLINK_H, POS_BLINK_X, POS_BLINK_Y);
}
void blink_nodisp()
{
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    blinkx[1], blinky[1], LED_BLINK_W, LED_BLINK_H, POS_BLINK_X, POS_BLINK_Y);
}
void InsertLCD()
{
  char buf[128];
  char driver_version[128];
  int apm_info_bios_major, apm_info_bios_minor, apm_info_bios_flags;
  int ac_line_status;
  int battery_status, battery_flag, percentage;
  int time_units;
  char units[128];
  FILE* fp;
  int i, j, n, x;

  fp = fopen(APM_FILE,"r");
  if (fp == NULL) {
    perror(APM_FILE);
    exit(1);
  }

  if (fgets(buf, sizeof(buf), fp) == NULL) {
    perror("");
    fclose(fp);
    return;
  }
  n = sscanf(buf,"%s %d.%d 0x%x 0x%x 0x%x 0x%x %d%% %d %s",
    driver_version,
    &apm_info_bios_major,
    &apm_info_bios_minor,
    &apm_info_bios_flags,
    &ac_line_status,
    &battery_status,
    &battery_flag,
    &percentage,
    &time_units,
    units);
  fclose(fp);
  
  if (n != 10) {
    fprintf(stderr, "not readed\n");
    return;
  }
#ifdef DEBUG
  printf("%s %d.%d 0x%x 0x%x 0x%x 0x%x %d%% %d %s\n",
    driver_version,
    apm_info_bios_major,
    apm_info_bios_minor,
    apm_info_bios_flags,
    ac_line_status,
    battery_status,
    battery_flag,
    percentage,
    time_units,
    units);
#endif

  if (battery_status == 0x03) { /* CHARGE */
    charge_on();
#ifdef DEBUG
    printf("CHARGE\n");
#endif
  } else {
    charge_off();
  }
  if (battery_flag != 0xff && (battery_flag & 0x80)) {
    battery_nodisp();
  } else {
    battery_disp();
  }
  if (mode == 0 || mode == 3) {
    blink_nodisp();
    if (ac_line_status == 0x01) { /* AC LINE */
      ac_disp();
#ifdef DEBUG
      printf("AC\n");
#endif
    } else {
      if (percentage > 0) {
        percent_disp();
        percentage = percentage/5*5; /* pc110 battery :-p */
        for (i = 3, x = 10; i >= 1; i--, x *= 10) {
          j = percentage%x/(x/10);
          if (i == 1 && j == 0) {
            XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
              ledx[10], ledy[10], LED_W, LED_H, posx[i], posy[i]);
          } else {
            XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
              ledx[j], ledy[j], LED_W, LED_H, posx[i], posy[i]);
          }
#ifdef DEBUG
          printf("j = %d\n", j);
#endif
        }
#ifdef DEBUG
        printf("percentage > 0\n");
#endif
      }
    }
  }
}
void InsertLCD2()
{
  time_t t;
  struct tm *clk;
  int idx;

  time(&t);
  clk = localtime(&t);  
  idx = clk->tm_hour/10;
  if (idx == 0) idx = 10;
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[idx], ledy[idx], LED_W, LED_H, posx[0], posy[0]);
  idx = clk->tm_hour%10;
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[idx], ledy[idx], LED_W, LED_H, posx[1], posy[1]);
  idx = clk->tm_min/10;
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[idx], ledy[idx], LED_W, LED_H, posx[2], posy[2]);
  idx = clk->tm_min%10;
  XCopyArea(dpy, led.pixmap, visible.pixmap, NormalGC,
    ledx[idx], ledy[idx], LED_W, LED_H, posx[3], posy[3]);
}
